'''
Game Manager
-Game Flow Control
-Game Objects Handling
'''

import pygame
from pygame.locals import *
import math
from random import randint

#Super class for game objects
class Material(pygame.sprite.DirtySprite):
    def __init__(self, loader):
        pygame.sprite.DirtySprite.__init__(self)
        self.loader = loader
        self.image = None
        self.rect = pygame.Rect((0, 0, 32, 32))
        self.area = None
        self.anime_count = 0
        self.anime_wait_max = 5
        self.anime_wait_count = 0
        self.anime_direction = 0

    def set_image(self, image_name, loader):
        #Not in use
        pass

    def move(self, old_pos, new_pos, target_list):
        angle = self.get_angle(old_pos, new_pos)
        dx = int( math.cos(angle * (math.pi / 180)) * self.speed )
        dy = int( math.sin(angle * (math.pi / 180)) * self.speed ) * -1
        hit_flag = False #never used default
        if dx != 0:
            hit_flag = self.move_sub(dx, 0, target_list)
        if dy != 0:
            hit_flag = self.move_sub(0, dy, target_list)

        self.anime_direction = self.get_direction(angle)
        return hit_flag

    def move_sub(self, dx, dy, target_list):
        self.rect.x += dx
        self.rect.y += dy
        hit_flag = False
        
        if self.rect.left < self.area.left:
            self.rect.left = self.area.left
            hit_flag = True
        if self.rect.right > self.area.right:
            self.rect.right = self.area.right
            hit_flag = True
        if self.rect.top < self.area.top:
            self.rect.top = self.area.top
            hit_flag = True
        if self.rect.bottom > self.area.bottom:
            self.rect.bottom = self.area.bottom
            hit_flag = True

        #Wall check
        my_hit_rect = self.get_hit_rect(self.rect, self.hit_rect_diff)
        for target in target_list:
            target_hit_rect = self.get_hit_rect(target.rect, target.hit_rect_diff)
            if my_hit_rect.colliderect(target_hit_rect):
                if dx > 0:
                    my_hit_rect.right = target_hit_rect.left
                    self.rect_update_sub(self.rect, my_hit_rect, self.hit_rect_diff)
                    hit_flag = True
                if dx < 0:
                    my_hit_rect.left = target_hit_rect.right
                    self.rect_update_sub(self.rect, my_hit_rect, self.hit_rect_diff)
                    hit_flag = True
                if dy > 0:
                    my_hit_rect.bottom = target_hit_rect.top
                    self.rect_update_sub(self.rect, my_hit_rect, self.hit_rect_diff)
                    hit_flag = True
                if dy < 0:
                    my_hit_rect.top = target_hit_rect.bottom
                    self.rect_update_sub(self.rect, my_hit_rect, self.hit_rect_diff)
                    hit_flag = True
        return hit_flag

    def rect_update_sub(self, target_rect, hit_rect, diff_rect):
        target_rect.topleft = (hit_rect.left-diff_rect[0], hit_rect.top-diff_rect[1])
        return

    def get_hit_rect(self, temp_rect, diff_rect):
        hit_rect = temp_rect.copy()
        hit_rect.topleft = (hit_rect.left+diff_rect[0], hit_rect.top+diff_rect[1])
        hit_rect.width = diff_rect[2]
        hit_rect.height = diff_rect[3]
        return hit_rect

    def get_direction(self, angle):
        if angle >= 45 and angle < 135:
            return 1
        elif angle >= 135 and angle < 225:
            return 2
        elif angle >= 225 and angle < 315:
            return 3
        else:
            return 0

    def get_angle(self, start_pos, end_pos):
        # Return 0-360 degree
        rad = math.atan2(start_pos[0] - end_pos[0], start_pos[1] - end_pos[1])
        angle = ((rad * (180.0 / math.pi)) + 90.0) % 360.0
        return angle

    def hit_check(self, target_list):
        my_hit_rect = self.get_hit_rect(self.rect, self.hit_rect_diff)
        target_hit_rect_list = []
        for target in target_list:
            target_hit_rect_list.append(self.get_hit_rect(target.rect, target.hit_rect_diff))
        return my_hit_rect.collidelistall(target_hit_rect_list)

    def animation(self, action, count):
        self.image = self.get_anime_image(action, count)
        self.anime_count_up(action)

    def get_anime_image(self, action, count):
        anime_list = self.loader.anime_dict[action]
        return anime_list[count+1]

    def anime_count_up(self, action):
        self.anime_wait_count += 1
        if self.anime_wait_count <= self.anime_wait_max:
            return
        self.anime_wait_count = 0
        anime_last = self.get_anime_last(action)
        self.anime_count += 1
        if self.anime_count >= anime_last:
            self.anime_count = 0

    def get_anime_last(self, action):
        anime_list = self.loader.anime_dict[action]
        return anime_list[0]

    def anime_random_start(self, name):
        if self.get_anime_last(name) == 1:
            return 0
        else:
            return randint(0,self.get_anime_last(name)-1)

    def update(self):
        self.dirty = 1

class Player(Material):
    def __init__(self, loader, screen):
        super(Player, self).__init__(loader)
        self.image = None
        self.hit_rect_diff = (8, 16, 16, 16)
        self.area = screen.get_rect()

        self.action = "standing"
        self.anime_stand = {0:"p_stand_right", 1:"p_stand_back", 2:"p_stand_left", 3:"p_stand_front"}
        self.anime_walk = {0:"p_walk_right", 1:"p_walk_back", 2:"p_walk_left", 3:"p_walk_front"}
        self.anime_jet = {0:"p_run_right", 1:"p_run_back", 2:"p_run_left", 3:"p_run_front"}
        self.anime_direction = 3
        self.stand()
        self.default_speed = 4
        self.speed = self.default_speed
        self.dizzy = 0
        self.jet_charge = 0
        self.jet_angle = 0
        self.jet_frame_count = 0
        self.crash_recovery = 0
        self.jet_number = 5 #Bad coding
  
        self.goal_flag = False

    def reset(self):
        self.action = "standing"
        self.anime_direction = 3
        self.stand()
        self.speed = self.default_speed
        self.dizzy = 0        
        self.jet_charge = 0
        self.jet_angle = 0
        self.jet_frame_count = 0
        self.jet_number = 5 #Bad coding

        self.goal_flag = False

    def stand(self):
        temp_anime_id = self.anime_stand[self.anime_direction]
        self.animation(temp_anime_id, self.anime_count)
        return 
        
    def direction_change(self):
        pass

    def jet_start_holding(self):
        self.action = "charging"
        self.jet_charge = 0

    def jet_holding(self):
        self.jet_charge += 1

    def jet_start(self):
        self.action = "jet_move"
        self.anime_count = 0
        old_pos = self.rect.topleft
        new_pos = pygame.mouse.get_pos()
        self.jet_angle = self.get_angle(old_pos, new_pos)
        self.jet_frame_count = 4
        if self.jet_charge > 30:
            self.jet_charge = 30
        self.anime_direction = self.get_direction(self.jet_angle)

    def jet_move(self, target_list):
        times = (self.jet_charge+1)*2
        early_hit = 0
        hit_flag = False
        for i in range(0, times):
            dx = int( math.cos(self.jet_angle * (math.pi / 180)) * self.speed )
            dy = int( math.sin(self.jet_angle * (math.pi / 180)) * self.speed ) * -1
            if dx != 0:
                hit_flag = self.move_sub(dx, 0, target_list)
            if dy != 0:
                hit_flag = self.move_sub(0, dy, target_list)
            if hit_flag == False:
                temp_anime_id = self.anime_jet[self.anime_direction]
                self.anime_count = 0
                self.animation(temp_anime_id, self.anime_count)
            else:
                early_hit = i
                self.jet_frame_count = 0
                break
        #JET ends
        self.jet_frame_count -= 1
        if self.jet_frame_count <= 0:
            if hit_flag == False:
                self.action = "sliding"
                self.jet_frame_count = int(((self.jet_charge+1) / 2))+1
                self.anime_count = 0
            else:
                if early_hit > 20:
                    self.start_crashed()
                else:
                    self.action = "standing"
                    self.anime_count = 0

    def start_crashed(self):
        self.action = "crashed"
        self.crash_recovery = 10
        self.loader.sound_dict["faint"].play()
        self.anime_count = 0
        self.animation("p_ko", self.anime_count)

    def crashed(self):
        self.animation("p_ko", self.anime_count)
        if self.crash_recovery <= 0:
            self.action = "standing"
            self.anime_count = 0
            self.crash_recovery = 0
        
    def jet_slide(self, target_list):
        hit_flag = False
        dx = int( math.cos(self.jet_angle * (math.pi / 180)) * (self.speed / (self.jet_frame_count + 1) ))
        dy = int( math.sin(self.jet_angle * (math.pi / 180)) * (self.speed / (self.jet_frame_count + 1) )) * -1
        if dx != 0:
            hit_flag = self.move_sub(dx, 0, target_list)
        if dy != 0:
            hit_flag = self.move_sub(0, dy, target_list)

        if hit_flag == True:
            self.jet_frame_count = 0
            if self.jet_frame_count >= 2:
                self.start_crashed()
            else:
                self.action = "standing"
                self.anime_count = 0
        else:
            self.jet_frame_count -= 1
            if self.jet_frame_count <= 0:
                if pygame.mouse.get_pressed()[0] == True:
                    self.start_running()
                else:
                    self.action = "standing"
                    self.anime_count = 0
                self.jet_charge = 0
                self.jet_frame_count = 0

    def start_running(self):
        self.action = "running"
        self.anime_count = 0
        self.anime_wait_count = 0

    def running(self, wall_list, item_list):
        old_pos = self.rect.topleft
        new_pos = pygame.mouse.get_pos()

        self.move((old_pos[0], old_pos[1]), new_pos, wall_list)
        temp_anime_id = self.anime_walk[self.anime_direction]
        self.animation(temp_anime_id, self.anime_count)

        sum_score = 0
        hit_list = self.hit_check(item_list)
        if hit_list != []:
            for id in hit_list:
                temp = item_list[id]
                if temp.item_name in ["coin_copper", "coin_silver", "coin_gold"]:
                    if temp.hit_on == True:
                        self.loader.sound_dict["coin"].play()
                        temp.visible = 0    #Workaround for not deleting objects
                        temp.hit_on = False #Workaround for not deleting objects
                        sum_score += temp.score_table[temp.item_name] #Bad code
                if temp.item_name == "exit":
                    self.goal_flag = True
                    self.loader.sound_dict["exit"].play()

        return sum_score  #Workaround

    def stop_running(self):
        pass

class Enemy(Material):
    def __init__(self, loader, screen):
        super(Enemy, self).__init__(loader)
        self.image = self.being_wall()
        self.hit_rect_diff = (0, 16, 32, 32)
        self.area = screen.get_rect()
        self.action = "wall"
        self.waiting_time = 0
        self.speed = 4
        
        self.anime_walk = {0:"e_walk_right", 1:"e_walk_back", 2:"e_walk_left", 3:"e_walk_front"}
        self.anime_tackle = {0:"e_tackle_right", 1:"e_tackle_back", 2:"e_tackle_left", 3:"e_tackle_front"}
        self.anime_direction = 3

    def being_wall(self):
        return self.get_anime_image("wall", 0)

    def start_braking_wall(self):
        #Wall particle here
        pass

    def braking_wall(self):
        sel

    def start_awakening(self):
        self.action = "awakening"
        self.waiting_time = randint(30,150)
        self.anime_count = 0 #rand
        self.hit_rect_diff = (4, 4, 24, 24) #adjust

    def awakening(self):
        self.animation("e_appear", self.anime_count)
        if randint(0,100) < 30: 
            self.image = pygame.transform.flip(self.image, 1, 0)
        self.waiting_time -= 1
        if self.waiting_time <= 0:
            self.action = "chasing"
            self.anime_count = 0 #rand

    def lying(self, flag):
        if flag == False:
            self.waiting_time -= 1
            if self.waiting_time <= 0:
                self.action = "chasing"
                self.anime_count = 0 #rand
            
    def chasing(self, player, enemy_list):
        old_pos = self.rect.topleft
        new_pos = player.rect.topleft

        dist = math.hypot(new_pos[0] - old_pos[0], new_pos[1] - old_pos[1])
        if dist > 200:
            self.speed = 2
        elif dist > 160:
            self.speed = 4
        elif dist < 60:
            self.speed = 5
            self.action = "start_tackling"
            return
        else:
            self.speed = 5

        #Remove the current enemy from the collision test
        temp_list = enemy_list[:]
        for i, t in enumerate(temp_list):
            if t is self:
                del temp_list[i]
                break

        self.move((old_pos[0], old_pos[1]), new_pos, temp_list)

        temp_anime_id = self.anime_walk[self.anime_direction]
        self.animation(temp_anime_id, self.anime_count)

    def start_tackling(self, player):
        self.action = "pausing"
        self.anime_count = 0
        self.speed = 8
        old_pos = self.rect.topleft
        new_pos = player.rect.topleft
        self.tackle_angle = self.get_angle(old_pos, new_pos)
        self.tackle_frame_count = 9
        self.anime_direction = self.get_direction(self.tackle_angle)
        self.waiting_time = 5

    def pausing(self):
        self.waiting_time -= 1
        if self.waiting_time <= 0:
            self.action = "tackling"
            self.image = self.get_anime_image(self.anime_tackle[self.anime_direction], 0)

    def tackle_move(self, player):
        dx = int( math.cos(self.tackle_angle * (math.pi / 180)) * self.speed )
        dy = int( math.sin(self.tackle_angle * (math.pi / 180)) * self.speed ) * -1
        if dx != 0:
            self.move_sub(dx, 0, [])
        if dy != 0:
            self.move_sub(0, dy, [])

        #Hit check
        enemy_hit_rect = self.get_hit_rect(self.rect, self.hit_rect_diff)
        player_hit_rect = player.get_hit_rect(player.rect, player.hit_rect_diff)
        if enemy_hit_rect.colliderect(player_hit_rect):
            self.action = "lying"
            self.loader.sound_dict["yelp"].play()
            self.takle_frame_count = 0
            return True

        #Tackle ends
        self.tackle_frame_count -= 1
        if self.tackle_frame_count <= 0:
            self.action = "lying"
            self.loader.sound_dict["thud"].play()
            self.waiting_time = 10
            self.anime_count = 0
        return False

class Coin(Material):
    def __init__(self, loader, coin_name):
        super(Coin, self).__init__(loader)
        self.item_name = coin_name
        self.hit_rect_diff = (8, 8, 16, 16)
        self.hit_on = True
        self.score_table = {"coin_copper":5, "coin_silver":20, "coin_gold":50}
        self.anime_count = self.anime_random_start(self.item_name)
        self.update()

    def update(self):
        super(Coin, self).update()
        self.animation(self.item_name, self.anime_count)

    def get_score(self):
        return self.score_table[self.item_name]

class Goal(Material):
    def __init__(self, loader):
        super(Goal, self).__init__(loader)
        self.item_name = "exit"
        self.hit_rect_diff = (0, 0, 32, 32)
        self.hit_on = True
        self.image = self.get_anime_image("exit", 0)
        
    def update(self):
        super(Goal, self).update()
        self.animation("exit", self.anime_count)

class Cursor(Material):
    def __init__(self, loader):
        super(Cursor, self).__init__(loader)
        self.image = self.get_anime_image("cursor", 0)
        self.image.set_alpha(100)

    def move(self):
        pos = pygame.mouse.get_pos()
        self.rect.topleft = pos

    def update(self):
        super(Cursor, self).update()
        self.animation("cursor", self.anime_count)

class Timer(Material):
    def __init__(self, loader, screen, default_time):
        super(Timer, self).__init__(loader)
        self.default_time = default_time * 1000
        self.play_time = 0.0
        self.zero_flag = False
        self.font = loader.font_counter #Bad coding
        self.area = screen.get_rect()
        self.show_time()

    def reset(self):
        self.play_time = 0.0
        self.zero_flag = False
        self.show_time()

    def show_time(self):
        self.image = pygame.Surface((self.area.width, self.area.height))#(size.get_width(), size.get_height()))
        self.image.set_colorkey((255, 255, 255))
        self.image = self.image.convert()
        self.image.fill((255, 255, 255))

        if self.play_time >= self.default_time:
            display_time = "00:00"
            text_color = (255, 32, 32)
        else:
            temp_msec = int(self.default_time - self.play_time)
            sec = int(temp_msec/1000.0)
            msec = int((temp_msec)%1000.0)
            display_time = ("0"+str(sec))[-2:]+":"+(str(msec))[:2]
            if self.play_time > self.default_time/2:
                text_color = (128, 96, 32)
            else:
                text_color = (16, 16, 16)

        text = self.font.render(display_time, 1, text_color)
        self.image.blit(text, (0,0))
        self.image = pygame.transform.scale(self.image, (self.area.width, int(self.area.height*1.8)))
        self.image.set_alpha(64)

        self.rect = self.image.get_rect()
        self.rect.topleft = 16, 92

    def add_time(self, msec):
        self.play_time += msec

    def start(self):
        pass

    def count_check(self):
        if self.play_time >= self.default_time:
            if self.zero_flag == False:
                self.zero_flag = True
                #Some Event for Time Zero
                return True
        return False

class Score:
    def __init__(self):
        #Gave up
        pass

class Life:
    def __init__(self, default_life):
        self.life_stock = default_life
        self.default_life = default_life

    def reset(self):
        self.life_stock = self.default_life

    def subtract(self):
        self.life_stock -= 1
        if self.life_stock <= 0:
            return 1
        return 0

class Jet(Material):
    def __init__(self, loader, screen, jet_stock):
        super(Jet, self).__init__(loader)
        self.font = loader.font_jet #Bad coding
        self.area = screen.get_rect()
        self.image = pygame.Surface((300,300))#(self.area.width, self.area.height))#(size.get_width(), size.get_height()))
        self.image.set_colorkey((255, 255, 255))
        self.image = self.image.convert()

        self.change(jet_stock)
        self.rect = self.image.get_rect()
        self.rect.topleft = 800, 704

    def change(self, jet_stock):
        self.image.fill((255, 255, 255))

        display_text = "JET "+str(jet_stock)
        text_color = (32, 64, 96)
        text = self.font.render(display_text, 1, text_color)
        self.image.blit(text, (0,0))
        
        #self.image = pygame.transform.scale(self.image, (int(self.area.width*0.6), int(self.area.height*0.6)))
        self.image.set_alpha(160)

class Stage(Material):
    def __init__(self, loader, screen, stage, life, score):
        super(Stage, self).__init__(loader)
        self.font = loader.font_stage #Bad coding
        self.area = screen.get_rect()
        self.image = pygame.Surface((self.area.width, self.area.height))#(size.get_width(), size.get_height()))
        self.image.set_colorkey((255, 255, 255))
        self.image = self.image.convert()
        self.image.fill((255, 255, 255))
        
        display_text = "STAGE"+str(stage)
        text_color = (32, 96, 64)
        text = self.font.render(display_text, 1, text_color)
        self.image.blit(text, (0,0))

        display_text = "LIFE "+str(life)
        text_color = (32, 96, 64)
        text = self.font.render(display_text, 1, text_color)
        self.image.blit(text, (0,224))

        self.image = pygame.transform.scale(self.image, (int(self.area.width*1.6), int(self.area.height*0.8)))
        self.image.set_alpha(192)

        self.rect = self.image.get_rect()
        self.rect.topleft = 92, 256

class GameManager:
    def __init__(self, screen, fps, loader, last_stage, default_life, default_time):
        self.loader = loader
        self.screen = screen
        self.high_score = self.loader.get_high_score()
        self.score = 0
        self.stage = 1
        self.last_stage = last_stage
        self.default_life = default_life
        self.default_time = default_time

        self.fps = fps
        self.background = None

        self.clock = pygame.time.Clock()

        self.timer_obj = Timer(self.loader, self.screen, self.default_time)
        self.score_obj = Score()
        self.life_obj = Life(self.default_life)

        self.player = Player(self.loader, self.screen)
        self.jet_obj = Jet(self.loader,  self.screen, self.player.jet_number) #Bad Coding

        self.cursor = Cursor(self.loader)
        self.enemy_list = []
        self.item_list = []

        self.allsprites = []
        
        self.game_end_flag = False
        self.game_end_time = 30

    def reset(self):
        self.score = 0
        self.stage = 1
        self.enemy_list = []
        self.item_list = []
        self.player.reset()
        self.allsprites = []
        self.timer_obj.reset()

        self.life_obj.reset()
        
        self.game_end_flag = False
        self.game_end_time = 30

    def restart(self):
        self.enemy_list = []
        self.item_list = []
        self.allsprites = []
        self.player.reset()
        self.timer_obj.reset()

        self.game_end_flag = False
        self.game_end_time = 30

        return self.start()
         
    def set_high_score(self, loaded_score):
        self.high_score = loaded_score
        
    def update_high_score(self):
        if self.score > self.high_score:
            self.high_score = self.score
            return 1
        return 0

    def get_screen(self):
        return self.screen

    def stage_clear(self):
        pass

    def goal_check(self):
        return 0

    def miss_check(self):
        return 0

    def change_layer_order(self):
        self.allsprites.change_layer(self.timer_obj, 0)
        #Sort by Y coordinate in rect()
        if self.game_end_flag == True:
            temp_list = []
            self.allsprites.change_layer(self.player, 0)
        else:
            temp_list = [self.player]
        temp_list = temp_list + self.enemy_list + self.item_list
        temp_list.sort(key = lambda x: x.rect[1])
        #Assign layer number. Lager is more prior.
        for layer_number, dirty_obj in enumerate(temp_list, start=1):
            self.allsprites.change_layer(dirty_obj, layer_number)
        self.allsprites.change_layer(self.cursor, layer_number+1)

    def map_setup(self):
        self.background = pygame.Surface(self.screen.get_size())
        self.background = self.background.convert()
        self.background.fill((255, 255, 255))
        
        stage_map = self.loader.map_list[self.stage-1]
        for y in range(0, 24):
            for x in range(0, 32):
                chip_num = stage_map[y][x]
                if chip_num == -1:
                    pass
                else:
                    img = self.loader.get_map_image(chip_num, self.stage)
                    self.background.blit(img, (x*32, y*32))

                    #Game object assignment
                    ret = self.loader.get_map_role(chip_num, self.stage)
                    if ret == "entrance":
                        self.player.rect.topleft = x*32, y*32
                    elif ret == "wall":
                        temp_enemy = Enemy(self.loader, self.screen)
                        self.enemy_list.append(temp_enemy)
                        temp_enemy.rect.topleft = x*32, y*32-16
                        #Change the block to the floor panel as a workaround
                        img = self.loader.get_floor_image()
                        for i in range(0, 2):
                            self.background.blit(img, (x*32, y*32-i*32))
                        #For debug
                        #temp_enemy.start_awakening()
                    elif ret in ["coin_copper", "coin_silver", "coin_gold"]:
                        temp_coin = Coin(self.loader, ret)
                        self.item_list.append(temp_coin)
                        temp_coin.rect.topleft = x*32, y*32-16
                        #Change the tile to the floor panel as a workaround
                        img = self.loader.get_floor_image()
                        self.background.blit(img, (x*32, y*32))
                    elif ret == "exit":
                        temp = Goal(self.loader)
                        self.item_list.append(temp)
                        temp.rect.topleft = x*32, y*32

    def start(self):
        self.map_setup()

        #Stage start screen
        stage_obj = Stage(self.loader, self.screen, self.stage, self.life_obj.life_stock, self.score)
        self.allsprites = pygame.sprite.LayeredDirty(self.player, self.enemy_list, self.item_list, self.cursor, stage_obj)
        self.allsprites.clear(self.screen, self.background)

        self.allsprites.change_layer(stage_obj, 99)
        
        #Gave up playing BGM

        total_msec = 0
        going = True
        abort = False
        while going:
            msec = self.clock.tick(self.fps)
            total_msec += msec
            if total_msec > 2000:
                going = False

            for event in pygame.event.get():
                if event.type == QUIT:
                    going = False
                    abort = True
                elif event.type == KEYDOWN and event.key == K_ESCAPE:
                    going = False
                    abort = True
        
            #self.change_layer_order()
            self.allsprites.update()
            rects = self.allsprites.draw(self.screen)
            pygame.display.update(rects)

        #Initialize
        self.timer_obj.start() #Not in use

        self.allsprites = pygame.sprite.LayeredDirty(self.player, self.enemy_list, self.item_list, self.cursor, self.timer_obj, self.jet_obj)
        self.allsprites.clear(self.screen, self.background)

        #Last stage event
        if self.stage == self.last_stage:
            for temp_enemy in self.enemy_list:
                temp_enemy.start_awakening()
            self.player.jet_number = 0
            self.jet_obj.change(self.player.jet_number)

        #Bad coding..
        if abort == False:
            ret = self.run()

            return ret
        else:
            return -1

    def run(self):

        going = True
        while going:
            msec = self.clock.tick(self.fps)
            self.timer_obj.add_time(msec)
            self.timer_obj.show_time()

            for event in pygame.event.get():
                if event.type == QUIT:
                    going = False
                elif event.type == KEYDOWN and event.key == K_ESCAPE:
                    going = False
                elif self.game_end_flag == True:
                    pass
                elif event.type == MOUSEBUTTONDOWN:
                    if pygame.mouse.get_pressed()[2] == True or pygame.mouse.get_pressed()[1] == True:
                        if self.player.jet_number >= 1:
                            self.player.jet_start_holding()
                    elif pygame.mouse.get_pressed()[0] == True:
                        if self.player.action in ["standing", "running"]:
                            self.player.start_running()                        
                elif event.type == MOUSEBUTTONUP:
                    #Stand or Releasing JET
                    if self.player.action == "charging":
                        if self.player.jet_number >= 1:
                            self.player.jet_number -= 1
                            self.jet_obj.change(self.player.jet_number) #Bad coding
                            self.loader.sound_dict["jet"].play()
                            self.player.jet_start()
                    if self.player.action == "running" and pygame.mouse.get_pressed()[0] != True:
                        #self.player.stop_running()
                        self.player.action = "standing"
                        self.player.stand()
                    if self.player.action == "crashed":
                        self.player.crash_recovery -= 1
                elif event.type == MOUSEMOTION:
                    self.cursor.move()
                    self.player.direction_change()

            if self.game_end_flag == False:
                if self.player.action == "standing":
                    self.player.stand()

                if self.player.action == "running":
                    sum_score = self.player.running(self.enemy_list, self.item_list)
                    self.score += sum_score
    
                if self.player.action == "charging":
                    self.player.jet_holding()
    
                if self.player.action == "jet_move":
                    self.player.jet_move(self.enemy_list)
    
                if self.player.action == "sliding":
                    self.player.jet_slide(self.enemy_list)
    
                if self.player.action == "crashed":
                    self.player.crashed()

            for temp_enemy in self.enemy_list:
                if temp_enemy.action == "awakening":
                    temp_enemy.awakening()
                if temp_enemy.action == "chasing":
                    temp_enemy.chasing(self.player, self.enemy_list)
                if temp_enemy.action == "start_tackling":
                    temp_enemy.start_tackling(self.player)
                if temp_enemy.action == "pausing":
                    temp_enemy.pausing()
                if temp_enemy.action == "lying":
                    temp_enemy.lying(self.game_end_flag)
                if temp_enemy.action == "tackling":
                    ret = temp_enemy.tackle_move(self.player)
                    if ret == True:
                        self.game_end_flag = True
                        self.game_end_time = self.timer_obj.play_time
                        #Player down animation
                        self.player.image = self.player.get_anime_image("p_down", 0)

            if self.timer_obj.count_check() == True:
                #Timer 0 event
                for temp_enemy in self.enemy_list:
                    temp_enemy.start_awakening()

            if self.player.goal_flag == True:
                self.stage += 1
                print ("Clear")
                if self.stage > self.last_stage:
                    #Ending
                    return 2
                #Next stage
                return 0

            if self.game_end_flag == True:
                if self.timer_obj.play_time - self.game_end_time >= 600:
                    #Gave up failure event
                    if self.life_obj.subtract() == 0:
                        #1 miss
                        return 0
                    else:
                        #Game Over
                        return 1

            self.change_layer_order()
            self.allsprites.update()
            rects = self.allsprites.draw(self.screen)
            pygame.display.update(rects)
        #Quit
        return -1
